print("import: 'js2py'")
import js2py

print("import: 'js2py.constructors'")
import js2py.constructors

print("import: 'js2py.es6'")
import js2py.es6

print("import: 'js2py.host'")
import js2py.host

print("import: 'js2py.internals'")
import js2py.internals

print("import: 'js2py.internals.constructors'")
import js2py.internals.constructors

print("import: 'js2py.internals.prototypes'")
import js2py.internals.prototypes

print("import: 'js2py.prototypes'")
import js2py.prototypes

print("import: 'js2py.py_node_modules'")
import js2py.py_node_modules

print("import: 'js2py.translators'")
import js2py.translators

print("import: 'js2py.utils'")
import js2py.utils

